@@ -5,6 +5,8 @@
 #include "local.h"
 #include "alarm.h"
 #include "filter.h"
+#include "channel.h"
+#include "internal.h"
 
 #include <stdint.h>
 #include <stdio.h>
@@ -12,27 +14,41 @@
 #include <syslog.h>
 
 /* UNIX facilities */
-#define FLUSH_MASTER	0	/* flush kernel conntrack table 	*/
-#define RESYNC_MASTER	1	/* resync with kernel conntrack table 	*/
-#define DUMP_INTERNAL 	16	/* dump internal cache 			*/
-#define DUMP_EXTERNAL 	17	/* dump external cache 			*/
-#define COMMIT		18	/* commit external cache		*/
-#define FLUSH_CACHE	19	/* flush cache				*/
-#define KILL		20	/* kill conntrackd			*/
-#define STATS		21	/* dump statistics			*/
-#define SEND_BULK	22	/* send a bulk				*/
-#define REQUEST_DUMP	23	/* request dump 			*/
-#define DUMP_INT_XML	24	/* dump internal cache in XML		*/
-#define DUMP_EXT_XML	25	/* dump external cache in XML		*/
-#define RESET_TIMERS	26	/* reset kernel timers			*/
-#define DEBUG_INFO	27	/* unused				*/
-#define STATS_NETWORK	28	/* extended network stats		*/
-#define STATS_CACHE	29	/* extended cache stats			*/
-#define STATS_RUNTIME	30	/* extended runtime stats		*/
-#define STATS_MULTICAST	31	/* multicast network stats		*/
-#define STATS_QUEUE	32	/* queue stats				*/
-#define FLUSH_INT_CACHE	33	/* flush internal cache			*/
-#define FLUSH_EXT_CACHE	34	/* flush external cache			*/
+#define CT_FLUSH_MASTER		0	/* flush kernel conntrack table */
+#define CT_RESYNC_MASTER	1	/* resync with kernel ct table	*/
+#define CT_DUMP_INTERNAL 	16	/* dump internal cache 		*/
+#define CT_DUMP_EXTERNAL 	17	/* dump external cache 		*/
+#define CT_COMMIT		18	/* commit external cache	*/
+#define CT_FLUSH_CACHE		19	/* flush cache			*/
+#define KILL			20	/* kill conntrackd		*/
+#define STATS			21	/* dump statistics		*/
+#define SEND_BULK		22	/* send a bulk			*/
+#define REQUEST_DUMP		23	/* request dump 		*/
+#define CT_DUMP_INT_XML		24	/* dump internal cache in XML	*/
+#define CT_DUMP_EXT_XML		25	/* dump external cache in XML	*/
+#define RESET_TIMERS		26	/* reset kernel timers		*/
+#define DEBUG_INFO		27	/* unused			*/
+#define STATS_NETWORK		28	/* extended network stats	*/
+#define STATS_CACHE		29	/* extended cache stats		*/
+#define STATS_RUNTIME		30	/* extended runtime stats	*/
+#define STATS_LINK		31	/* dedicated link stats		*/
+#define STATS_RSQUEUE		32	/* resend queue stats		*/
+#define CT_FLUSH_INT_CACHE	33	/* flush internal cache		*/
+#define CT_FLUSH_EXT_CACHE	34	/* flush external cache		*/
+#define STATS_PROCESS		35	/* child process stats		*/
+#define STATS_QUEUE		36	/* queue stats			*/
+#define EXP_STATS		37	/* dump statistics		*/
+#define EXP_FLUSH_MASTER	38	/* flush kernel expect table    */
+#define EXP_RESYNC_MASTER	39	/* resync with kernel exp table	*/
+#define EXP_DUMP_INTERNAL	40	/* dump internal expect cache	*/
+#define EXP_DUMP_EXTERNAL	41	/* dump external expect cache	*/
+#define EXP_COMMIT		42	/* commit expectations		*/
+#define ALL_FLUSH_MASTER	43	/* flush all kernel tables	*/
+#define ALL_RESYNC_MASTER	44	/* resync w/all kernel tables	*/
+#define ALL_FLUSH_CACHE		45	/* flush all caches		*/
+#define ALL_COMMIT		46	/* commit all tables		*/
+#define EXP_DUMP_INT_XML	47	/* dump internal cache in XML	*/
+#define EXP_DUMP_EXT_XML	48	/* dump external cache in XML	*/
 
 #define DEFAULT_CONFIGFILE	"/etc/conntrackd/conntrackd.conf"
 #define DEFAULT_LOCKFILE	"/var/lock/conntrackd.lock"
@@ -52,6 +68,8 @@
 #define CTD_SYNC_ALARM		(1UL << 3)
 #define CTD_SYNC_NOTRACK	(1UL << 4)
 #define CTD_POLL		(1UL << 5)
+#define CTD_EXPECT		(1UL << 6)
+#define CTD_HELPER		(1UL << 7)
 
 /* FILENAME_MAX is 4096 on my system, perhaps too much? */
 #ifndef FILENAME_MAXLEN
@@ -71,11 +89,11 @@ struct ct_conf {
 	int syslog_facility;
 	char lockfile[FILENAME_MAXLEN];
 	int hashsize;			/* hashtable size */
-	int mcast_links;
-	int mcast_default_link;
-	struct mcast_conf mcast[MCAST_LINKS_MAX];
+	int channel_num;
+	int channel_default;
+	int channel_type_global;
+	struct channel_conf channel[MULTICHANNEL_MAX];
 	struct local_conf local;	/* unix socket facilities */
-	int nice;
 	int limit;
 	int refresh;
 	int cache_timeout;		/* cache entries timeout */
@@ -84,21 +102,43 @@ struct ct_conf {
 	unsigned int netlink_buffer_size;
 	unsigned int netlink_buffer_size_max_grown;
 	int nl_overrun_resync;
-	union inet_address *listen_to;
-	unsigned int listen_to_len;
 	unsigned int flags;
-	int family;			/* protocol family */
 	unsigned int resend_queue_size; /* FTFW protocol */
 	unsigned int window_size;
 	int poll_kernel_secs;
-	int cache_write_through;
 	int filter_from_kernelspace;
 	int event_iterations_limit;
+	int systemd;
+	int running_mode;
+	int startup_resync;
+	struct {
+		int error_queue_length;
+	} channelc;
+	struct {
+		int internal_cache_disable;
+		int external_cache_disable;
+		int tcp_window_tracking;
+	} sync;
+	struct {
+		int subsys_id;
+		int groups;
+		int events_reliable;
+	} netlink;
+	struct {
+		int commit_steps;
+	} general;
+	struct {
+		int type;
+		int prio;
+	} sched;
 	struct {
 		char logfile[FILENAME_MAXLEN];
 		int syslog_facility;
 		size_t buffer_size;
 	} stats;
+	struct {
+		struct list_head list;
+	} cthelper;
 };
 
 #define STATE(x) st.x
@@ -107,19 +147,21 @@ struct ct_general_state {
 	sigset_t 			block;
 	FILE 				*log;
 	FILE				*stats_log;
+	int				log_init;
 	struct local_server		local;
 	struct ct_mode 			*mode;
 	struct ct_filter		*us_filter;
+	struct exp_filter		*exp_filter;
 
 	struct nfct_handle		*event;         /* event handler */
 	struct nfct_filter		*filter;	/* event filter */
 	int				event_iterations_limit;
 
 	struct nfct_handle		*dump;		/* dump handler */
-	struct nfct_handle		*request;	/* request handler */
 	struct nfct_handle		*resync;	/* resync handler */
 	struct nfct_handle		*get;		/* get handler */
 	int				get_retval;	/* hackish */
+	struct nfct_handle		*flush;		/* flusher */
 
 	struct alarm_block		resync_alarm;
 	struct alarm_block		polling_alarm;
@@ -157,17 +199,37 @@ struct ct_general_state {
 	} stats;
 };
 
+struct commit_runqueue {
+	int		(*cb)(struct nfct_handle *h, int step);
+};
+
 #define STATE_SYNC(x) state.sync->x
 
 struct ct_sync_state {
-	struct cache *internal; 	/* internal events cache (netlink) */
-	struct cache *external; 	/* external events cache (mcast) */
+	struct external_handler *external;
 
-	struct mcast_sock_multi *mcast_server;  /* multicast incoming */
-	struct mcast_sock_multi *mcast_client;  /* multicast outgoing  */
-	struct nlif_handle	*mcast_iface;
+	struct multichannel	*channel;
+	struct nlif_handle	*interface;
 	struct queue *tx_queue;
 
+#define COMMIT_STATE_INACTIVE	0
+#define COMMIT_STATE_MASTER	1
+#define COMMIT_STATE_RELATED	2
+
+	struct {
+		int			state;
+		int			clientfd;
+		struct nfct_handle	*h;
+		struct evfd		*evfd;
+		int			current;
+		struct commit_runqueue  rq[2];
+		struct {
+			int 		ok;
+			int		fail;
+			struct timeval	start;
+		} stats;
+	} commit;
+
 	struct alarm_block		reset_cache_alarm;
 
 	struct sync_mode *sync;		/* sync mode */
@@ -196,37 +258,41 @@ struct ct_stats_state {
 	struct cache *cache;            /* internal events cache (netlink) */
 };
 
-union ct_state {
+#define STATE_CTH(x) state.cthelper->x
+
+struct ct_helper_state {
+	struct mnl_socket *nl;
+	uint32_t portid;
+};
+
+struct ct_state {
 	struct ct_sync_state *sync;
 	struct ct_stats_state *stats;
+	struct ct_helper_state *cthelper;
 };
 
 extern struct ct_conf conf;
-extern union ct_state state;
+extern struct ct_state state;
 extern struct ct_general_state st;
 
-#ifndef IPPROTO_VRRP
-#define IPPROTO_VRRP 112
-#endif
-
-#define STEPS_PER_SECONDS	5
-
 struct ct_mode {
+	struct internal_handler *internal;
 	int (*init)(void);
-	void (*run)(fd_set *readfds);
 	int (*local)(int fd, int type, void *data);
 	void (*kill)(void);
-	void (*dump)(struct nf_conntrack *ct);
-	int (*resync)(enum nf_conntrack_msg_type type,
-		      struct nf_conntrack *ct,
-		      void *data);
-	int (*purge)(void);
-	void (*event_new)(struct nf_conntrack *ct);
-	void (*event_upd)(struct nf_conntrack *ct);
-	int (*event_dst)(struct nf_conntrack *ct);
 };
 
-/* conntrackd modes */
+/* basic ctnl functions */
+void ctnl_kill(void);
+int ctnl_local(int fd, int type, void *data);
+int ctnl_init(void);
+
+/* basic cthelper functions */
+void cthelper_kill(void);
+int cthelper_local(int fd, int type, void *data);
+int cthelper_init(void);
+
+/* conntrackd ctnl modes */
 extern struct ct_mode sync_mode;
 extern struct ct_mode stats_mode;
 
@@ -234,12 +300,16 @@ extern struct ct_mode stats_mode;
 
 /* These live in run.c */
 void killer(int foo);
-void local_handler(int fd, void *data);
+int evaluate(void);
 int init(void);
-void run(void);
+void select_main_loop(void);
 
 /* from read_config_yy.c */
 int
 init_config(char *filename);
 
+#ifndef ARRAY_SIZE
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
+#endif
+
 #endif
